import time
from threading import Semaphore
from typing import Optional

import numpy as np

from dt_modeling.odometry.types import Pose2DEstimate


class VelocityToPose:
    """
    `VelocityToPose` integrates the velocity of the robot over time in order to continuously
    obtain a pose estimate that is relative to the pose at which the integration started.

    """
    def __init__(self):
        # utility objects
        self._lock: Semaphore = Semaphore()
        # internal state
        self._last_theta_dot: float = 0
        self._last_v: float = 0
        self._last_timestamp: Optional[float] = None
        self._has_estimate: bool = False
        # current estimate
        self._pose: Pose2DEstimate = Pose2DEstimate(0, 0, 0, 0)

    def update(self, v: float, omega: float, timestamp: float = None):
        """
        Performs the calclulation from velocity to pose and publishes a messsage with the result.

        Args:
            v (:obj:`float`):           desired linear velocity of the chassis in meters/second
            omega (:obj:`float`):       desired angular velocity of the chassis in radians/second
            timestamp (:obj:`float`):   timestamp in seconds when the reading was performed, if
                                        None is given, the current time is used

        """
        timestamp = timestamp if timestamp is not None else time.time()

        with self._lock:
            if self._last_timestamp is not None:
                # compute delta_t
                dt = timestamp - self._last_timestamp

                # integrate the relative movement between the last pose and the current
                theta_delta = self._last_theta_dot * dt

                # to ensure no division by zero for radius calculation:
                if np.abs(self._last_theta_dot) < 0.000001:
                    # straight line
                    x_delta = self._last_v * dt
                    y_delta = 0
                else:
                    # arc of circle
                    radius = self._last_v / self._last_theta_dot
                    x_delta = radius * np.sin(theta_delta)
                    y_delta = radius * (1.0 - np.cos(theta_delta))

                # add to the previous to get absolute pose relative to the starting position
                theta_res = self._pose.theta + theta_delta
                x_res = self._pose.x + \
                        x_delta * np.cos(self._pose.theta) - \
                        y_delta * np.sin(self._pose.theta)
                y_res = self._pose.y + \
                        y_delta * np.cos(self._pose.theta) + \
                        x_delta * np.sin(self._pose.theta)

                print(dt, x_delta, x_res)

                # update the stored pose
                self._pose.theta = theta_res
                self._pose.x = x_res
                self._pose.y = y_res
                self._pose.time = timestamp
                self._has_estimate = True

            # update internal state
            self._last_timestamp = timestamp
            self._last_theta_dot = omega
            self._last_v = v

    def get_estimate(self) -> Optional[Pose2DEstimate]:
        with self._lock:
            if not self._has_estimate:
                return None
            return self._pose.copy()



import dataclasses

import numpy as np

import transformations as tr


@dataclasses.dataclass
class Pose2DEstimate:
    x: float
    y: float
    theta: float
    # time of the estimate
    time: float

    @property
    def q(self) -> np.ndarray:
        return tr.quaternion_from_euler(0, 0, self.theta)

    def copy(self) -> 'Pose2DEstimate':
        return Pose2DEstimate(**dataclasses.asdict(self))

    def __str__(self):
        return f"P(x={self.x}, y={self.y}, theta={self.theta})"



from typing import Dict, Tuple

from dt_modeling.kinematics.utils import trim_value


class PWM:
    """
    The `PWM` maps wheels' speed to duty cycle commands to send to the DC motors.
    It calculates the PWM commands that the DC motors should execute in order for the robot to
    perform a desired wheels' command.

    Args:
        gain (:obj:`float`):            scaling factor applied to the desired velocity
        trim (:obj:`float`):            trimming factor used to offset differences in the
                                        behaviour of the left and right motors, it is recommended
                                        to use a value that results in the robot moving in a
                                        straight line when equal forward commands are given
        k (:obj:`float`):               motor constant, assumed equal for both motors
        limit (:obj:`float`):           limits the final commands sent to the motors

    """

    def __init__(self,
                 gain: float = 1.0,
                 trim: float = 0.0,
                 k: float = 27.0,
                 limit: float = 1.0,
                 ):
        # store parameters
        self.gain: float = gain
        self.trim: float = trim
        self.k: float = k
        self.limit: float = limit

    def get_current_configuration(self) -> Dict[str, float]:
        return {
            "gain": self.gain,
            "trim": self.trim,
            "k": self.k,
            "limit": self.limit,
        }

    def get_wheels_gain(self) -> Tuple[float, float]:
        """
        Returns the wheels' gain adjusted by gain and trim.

        Returns:
            (:obj:`float`):         left wheel's gain
            (:obj:`float`):         right wheel's gain

        """
        # assuming same motor constants k for both motors
        k_r = k_l = self.k

        # adjusting k by gain and trim
        k_l_inv = (self.gain - self.trim) / k_l
        k_r_inv = (self.gain + self.trim) / k_r

        return k_l_inv, k_r_inv

    def get_wheels_duty_cycle(self, omega_l: float, omega_r: float) -> \
            Tuple[float, float]:
        """
        Maps the given wheel speeds to duty cycle commands that the robot can execute directly.

        Args:
            (:obj:`float`):         rotation speed of the left wheel in radians/second
            (:obj:`float`):         rotation speed of the right wheel in radians/second

        Returns:
            (:obj:`float`):         command to be sent to the left wheel
            (:obj:`float`):         command to be sent to the right wheel

        """
        # adjusting k by gain and trim
        k_l_inv, k_r_inv = self.get_wheels_gain()

        # conversion from motor rotation rate to duty cycle
        u_r = omega_r * k_r_inv
        u_l = omega_l * k_l_inv

        # limiting output to limit, which is 1.0 for the duckiebot
        u_r_limited = trim_value(u_r, -self.limit, self.limit)
        u_l_limited = trim_value(u_l, -self.limit, self.limit)

        return u_l_limited, u_r_limited
